@@ -3,7 +3,7 @@ import contextvars
 import inspect
 import logging
 from collections import ChainMap
-from contextlib import AsyncExitStack, AbstractAsyncContextManager, AbstractContextManager, asynccontextmanager
+from contextlib import AsyncExitStack, AbstractAsyncContextManager, asynccontextmanager
 from json import JSONDecodeError
 from types import FunctionType, CoroutineType
 from typing import List, Union, Any, Callable, Type, Optional, Dict, Sequence, Awaitable
@@ -486,21 +486,40 @@ def make_request_model(name, module, body_params: List[ModelField]):
 
 
 class JsonRpcContext:
-    def __init__(self, entrypoint: 'Entrypoint', raw_request: Any):
+    def __init__(
+        self,
+        entrypoint: 'Entrypoint',
+        raw_request: Any,
+        http_request: Request,
+        background_tasks: BackgroundTasks,
+        http_response: Response,
+    ):
         self.entrypoint: Entrypoint = entrypoint
         self.raw_request: Any = raw_request
-        self.raw_response: Optional[dict] = None
+        self.http_request: Request = http_request
+        self.background_tasks: BackgroundTasks = background_tasks
+        self.http_response: Response = http_response
+        self._raw_response: Optional[dict] = None
+        self.exception: Optional[Exception] = None
         self.exit_stack: Optional[AsyncExitStack] = None
         self.jsonrpc_context_token: Optional[contextvars.Token] = None
 
-    def on_raw_response(self, raw_response: dict):
+    def on_raw_response(self, raw_response: dict, exception: Optional[Exception] = None):
         raw_response.pop('id', None)
         if isinstance(self.raw_request, dict) and 'id' in self.raw_request:
             raw_response['id'] = self.raw_request.get('id')
         elif 'error' in raw_response:
             raw_response['id'] = None
-        self.raw_response = raw_response
-        return self.raw_response
+        self._raw_response = raw_response
+        self.exception = exception
+
+    @property
+    def raw_response(self) -> dict:
+        return self._raw_response
+
+    @raw_response.setter
+    def raw_response(self, value: dict):
+        self.on_raw_response(value)
 
     @cached_property
     def request(self) -> JsonRpcRequest:
@@ -518,7 +537,7 @@ class JsonRpcContext:
     async def __aenter__(self):
         assert self.exit_stack is None
         self.exit_stack = await AsyncExitStack().__aenter__()
-        await self.exit_stack.enter_async_context(self._handle_exception())
+        await self.exit_stack.enter_async_context(self._handle_exception(reraise=False))
         self.jsonrpc_context_token = _jsonrpc_context.set(self)
         return self
 
@@ -528,12 +547,21 @@ class JsonRpcContext:
         return await self.exit_stack.__aexit__(*exc_details)
 
     @asynccontextmanager
-    async def _handle_exception(self):
+    async def _handle_exception(self, reraise):
         try:
             yield
         except Exception as exc:
-            resp = await self.entrypoint.handle_exception_to_resp(exc)
-            self.on_raw_response(resp)
+            if exc is not self.exception:
+                try:
+                    resp = await self.entrypoint.handle_exception_to_resp(exc)
+                except Exception as exc:
+                    # HTTPException
+                    self._raw_response = None
+                    self.exception = exc
+                    raise
+                self.on_raw_response(resp, exc)
+            if reraise:
+                raise
 
     async def enter_middlewares(self, middlewares: Sequence['JsonRpcMiddleware']):
         for mw in middlewares:
@@ -541,7 +569,7 @@ class JsonRpcContext:
             if not isinstance(cm, AbstractAsyncContextManager):
                 raise RuntimeError("JsonRpcMiddleware(context) must return AsyncContextManager")
             await self.exit_stack.enter_async_context(cm)
-            await self.exit_stack.enter_async_context(self._handle_exception())
+            await self.exit_stack.enter_async_context(self._handle_exception(reraise=True))
 
 
 JsonRpcMiddleware = Callable[[JsonRpcContext], AbstractAsyncContextManager]
@@ -711,7 +739,13 @@ class MethodRoute(APIRoute):
         dependency_cache: dict = None,
         shared_dependencies_error: BaseError = None
     ) -> dict:
-        async with JsonRpcContext(self.entrypoint, req) as ctx:
+        async with JsonRpcContext(
+            entrypoint=self.entrypoint,
+            raw_request=req,
+            http_request=http_request,
+            background_tasks=background_tasks,
+            http_response=sub_response,
+        ) as ctx:
             await ctx.enter_middlewares(self.entrypoint.middlewares)
 
             if ctx.request.method != self.name:
@@ -1029,7 +1063,13 @@ class EntrypointRoute(APIRoute):
         dependency_cache: dict = None,
         shared_dependencies_error: BaseError = None
     ) -> dict:
-        async with JsonRpcContext(self.entrypoint, req) as ctx:
+        async with JsonRpcContext(
+            entrypoint=self.entrypoint,
+            raw_request=req,
+            http_request=http_request,
+            background_tasks=background_tasks,
+            http_response=sub_response,
+        ) as ctx:
             await ctx.enter_middlewares(self.entrypoint.middlewares)
 
             resp = await self.handle_req(
